'''Cloning design classes and functions.'''
from ._primers import primer, primers
from ._sequence_generation.random_sequences import random_dna
from ._sequence_generation.random_sequences import random_codons
from ._oligo_synthesis.oligo_assembly import OligoAssembly
from ._gibson import gibson_primers
from ._gibson import gibson
