'''Genbank constants'''

TO_CORAL = {'-10_signal': '-10_signal',
            '-35_signal': '-35_signal',
            '3\'UTR': '3\'UTR',
            '5\'UTR': '5\'UTR',
            'assembly_gap': 'assembly_gap',
            'attenuator': 'attenuator',
            'C_region': 'C_region',
            'CAAT_signal': 'CAAT_signal',
            'CDS': 'CDS',
            'centromere': 'centromere',
            'D-loop': 'D-loop',
            'D_segment': 'D_segment',
            'enhancer': 'enhancer',
            'exon': 'exon',
            'gap': 'gap',
            'GC_signal': 'GC_signal',
            'gene': 'gene',
            'iDNA': 'iDNA',
            'intron': 'intron',
            'J_segment': 'J_segment',
            'LTR': 'LTR',
            'mat_peptide': 'mat_peptide',
            'misc_binding': 'misc_binding',
            'misc_difference': 'misc_difference',
            'misc_feature': 'misc_feature',
            'misc_recomb': 'misc_recomb',
            'misc_RNA': 'misc_RNA',
            'misc_signal': 'misc_signal',
            'misc_structure': 'misc_structure',
            'mobile_element': 'mobile_element',
            'modified_base': 'modified_base',
            'mRNA': 'mRNA',
            'ncRNA': 'ncRNA',
            'N_region': 'N_region',
            'old_sequence': 'old_sequence',
            'operon': 'operon',
            'oritT': 'oriT',
            'polyA_signal': 'polyA_signal',
            'polyA_site': 'polyA_site',
            'precursor_RNA': 'precursor_RNA',
            'prim_transcript': 'prim_transcript',
            'primer_bind': 'primer_bind',
            'promoter': 'promoter',
            'protein_bind': 'protein_bind',
            'RBS': 'RBS',
            'repeat_region': 'repeat_region',
            'rep_origin': 'rep_origin',
            'rRNA': 'rRNA',
            'S_region': 'S_region',
            'sig_peptide': 'sig_peptide',
            'source': 'source',
            'stem_loop': 'stem_loop',
            'STS': 'STS',
            'TATA_signal': 'TATA_signal',
            'telomere': 'telomere',
            'terminator': 'terminator',
            'tmRNA': 'tmRNA',
            'transit_peptide': 'transit_peptide',
            'tRNA': 'tRNA',
            'unsure': 'unsure',
            'V_region': 'V_region',
            'V_segment': 'V_segment',
            'variation': 'variation'}
# 'site' appears in some ApE files / Geneious exports, not sure why
TO_CORAL['site'] = 'site'
TO_BIO = {value: key for key, value in TO_CORAL.iteritems()}
